<div class="modal fade bd-example-modal-lg" tabindex="-1" role="dialog" aria-labelledby="Settings" aria-hidden="true">
  <div class="modal-dialog modal-lg">
    <div class="modal-content">
	    <div class="modal-header">
	      <button type="button" class="close" data-dismiss="modal" aria-label="Close">
	        <span aria-hidden="true">&times;</span>
	      </button>
	      <h4 class="modal-title">Settings</h4>
	    </div>
	    <div class="modal-body">
	       <ul class="nav nav-tabs" id="tabContent">
	           <li class="nav-item">
	             <a class="active nav-link" href="#general" data-toggle="tab">General</a>
	           </li>
	           <li class="nav-item">
	             <a class="nav-link" href="#model" data-toggle="tab">Model</a>
	           </li>
           </ul>
            <div class="tab-content">
	            <div class="tab-pane active" id="general">
	               <div class="container" style="padding-top:15px;">
                       {% include "settings_general.html" %}
                   </div>
	            </div>
	        
	            <div class="tab-pane" id="model">
	               <div class="container" style="padding-top:15px;">
	                   {% include "settings_model.html" %}
	               </div>
	            </div> 
	 
	        </div>
	    </div>
	    <div class="modal-footer">
	      <button type="button" class="btn btn-secondary" data-dismiss="modal">Close</button>
	    </div>
    </div>
  </div>
</div>

