<div class="dropdown container">
 <button class="btn btn-secondary dropdown-toggle btn-block" type="button" data-toggle="dropdown">Lighting<span class="caret"></span></button>
  <ul class="dropdown-menu">
    <li onclick="toggle_dynamic_lighting(this, dynamic_lighting)" id="dynamic_lighting"><a class="dropdown-item" href="#">Dynamic Lighting</a></li>
    <li onclick="toggle_entity_shadows(this, entity_shadows)" id="entity_shadows"><a class="dropdown-item" href="#">Object Shadows</a></li>
    <li onclick="toggle_terrain_shadows(this, terrain_shadows)" id="terrain_shadows"><a class="dropdown-item" href="#">Terrain Shadows</a></li>
  </ul>
</div>
<div class="container" style="padding-top:15px;">
 <a class="btn btn-secondary btn-block" id="toggle_pos_target_line">Toggle Target Line</a>
</div>
<div class="container" style="padding-top:15px;">
 <a class="btn btn-secondary btn-block" id="toggle_fence">Toggle Fence</a>
</div>
<div class="container" style="padding-top:15px;">
 <a class="btn btn btn-outline-primary btn-block" id="send_auto">AUTO</a>
</div>
<div class="container" style="padding-top:15px;">
 <a class="btn btn-secondary btn-block" id="hud_width">Toggle HUD Width</a>
</div>
<div class="container" style="padding-top:15px;">
 <a class="btn btn-secondary btn-block" id="toggle_sensor_footprint" onclick="toggle_sensor_footprint(this, show_sensor_footprint)">Toggle Sensor Footprint (experimental)</a>
</div>
