<ul class="dropdown-menu multi-level" role="menu" aria-labelledby="dropdownMenu" style="display:block;position:static;margin-bottom:5px;">
	{% if markers %}
        {% for id in markers %}
            <li><a class="dropdown-item" tabindex="-1" href="#" id = {{ escape(id) }}>Set</a></li>
            <li><a class="dropdown-item" tabindex="-1" href="#" id = {{ escape(id) }}>Remove</a></li>
            <li><a class="dropdown-item" tabindex="-1" href="#" id = {{ escape(id) }}>Move</a></li>
        {% end %}
        <li class="dropdown-divider"></li>
    {% end %}
	<li class="dropdown-submenu">
        <a class="dropdown-item" tabindex="-1" href="#">Mission</a>
        <ul class="dropdown-menu">
	        <li><a class="dropdown-item" id ="mission-clear" href="#">Clear</a></li>
	        <li><a class="dropdown-item" id ="mission-list" href="#">List</a></li>
	        <li><a class="dropdown-item" id ="mission-load" href="#">Load</a></li>
        </ul>
    </li>
    <li class="dropdown-submenu">
        <a class="dropdown-item" tabindex="-1" href="#">Fence</a>
        <ul class="dropdown-menu">
            <li><a class="dropdown-item" id ="fence-clear" href="#">Clear</a></li>
            <li><a class="dropdown-item" id ="fence-list" href="#">List</a></li>
            <li><a class="dropdown-item" id ="fence-load" href="#">Load</a></li>
        </ul>
    </li>
    <li class="dropdown-divider"></li>
    <li><a class="dropdown-item" tabindex="-1" href="#">Show Landing Zone</a></li>
</ul>
